@@ -1,5 +1,5 @@
 /* df - summarize free disk space
-   Copyright (C) 1991, 1995-2010 Free Software Foundation, Inc.
+   Copyright (C) 1991-2012 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,17 +22,20 @@
 #include <stdio.h>
 #include <sys/types.h>
 #include <getopt.h>
+#include <assert.h>
 
 #include "system.h"
+#include "canonicalize.h"
 #include "error.h"
 #include "fsusage.h"
 #include "human.h"
+#include "mbsalign.h"
+#include "mbswidth.h"
 #include "mountlist.h"
 #include "quote.h"
-#include "save-cwd.h"
-#include "xgetcwd.h"
+#include "find-mount-point.h"
 
-/* The official name of this program (e.g., no `g' prefix).  */
+/* The official name of this program (e.g., no 'g' prefix).  */
 #define PROGRAM_NAME "df"
 
 #define AUTHORS \
@@ -66,7 +69,7 @@ static bool posix_format;
 /* True if a file system has been processed for output.  */
 static bool file_systems_processed;
 
-/* If true, invoke the `sync' system call before getting any usage data.
+/* If true, invoke the 'sync' system call before getting any usage data.
    Using this option can make df very slow, especially with many or very
    busy disks.  Note that this may make a difference on some systems --
    SunOS 4.1.3, for one.  It is *not* necessary on GNU/Linux.  */
@@ -84,7 +87,7 @@ struct fs_type_list
 };
 
 /* Linked list of file system types to display.
-   If `fs_select_list' is NULL, list all types.
+   If 'fs_select_list' is NULL, list all types.
    This table is generated dynamically from command-line options,
    rather than hardcoding into the program what it thinks are the
    valid file system types; let the user specify any file system type
@@ -113,12 +116,59 @@ static bool print_grand_total;
 /* Grand total data. */
 static struct fs_usage grand_fsu;
 
+/* Display modes.  */
+enum { DEFAULT_MODE, INODES_MODE, HUMAN_MODE, POSIX_MODE, NMODES };
+static int header_mode = DEFAULT_MODE;
+
+/* Displayable fields.  */
+enum
+{
+  DEV_FIELD,   /* file system */
+  TYPE_FIELD,  /* FS type */
+  TOTAL_FIELD, /* blocks or inodes */
+  USED_FIELD,  /* ditto */
+  FREE_FIELD,  /* ditto */
+  PCENT_FIELD, /* percent used */
+  MNT_FIELD,   /* mount point */
+  NFIELDS
+};
+
+/* Header strings for the above fields in each mode.
+   NULL means to use the header for the default mode.  */
+static const char *headers[NFIELDS][NMODES] = {
+/*  DEFAULT_MODE	INODES_MODE	HUMAN_MODE	POSIX_MODE  */
+  { N_("Filesystem"),   NULL,           NULL,           NULL },
+  { N_("Type"),         NULL,           NULL,           NULL },
+  { N_("blocks"),       N_("Inodes"),   N_("Size"),     NULL },
+  { N_("Used"),         N_("IUsed"),    NULL,           NULL },
+  { N_("Available"),    N_("IFree"),    N_("Avail"),    NULL },
+  { N_("Use%"),         N_("IUse%"),    NULL,           N_("Capacity") },
+  { N_("Mounted on"),   NULL,           NULL,           NULL }
+};
+
+/* Alignments for the 3 textual and 4 numeric fields.  */
+static mbs_align_t alignments[NFIELDS] = {
+  MBS_ALIGN_LEFT, MBS_ALIGN_LEFT,
+  MBS_ALIGN_RIGHT, MBS_ALIGN_RIGHT, MBS_ALIGN_RIGHT, MBS_ALIGN_RIGHT,
+  MBS_ALIGN_LEFT
+};
+
+/* Auto adjusted (up) widths used to align columns.  */
+static size_t widths[NFIELDS] = { 14, 4, 5, 5, 5, 4, 0 };
+
+/* Storage for pointers for each string (cell of table).  */
+static char ***table;
+
+/* The current number of processed rows (including header).  */
+static size_t nrows;
+
 /* For long options that have no equivalent short option, use a
    non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
 enum
 {
   NO_SYNC_OPTION = CHAR_MAX + 1,
-  SYNC_OPTION
+  SYNC_OPTION,
+  MEGABYTES_OPTION  /* FIXME: remove long opt in Aug 2013 */
 };
 
 static struct option const long_options[] =
@@ -129,7 +179,7 @@ static struct option const long_options[] =
   {"human-readable", no_argument, NULL, 'h'},
   {"si", no_argument, NULL, 'H'},
   {"local", no_argument, NULL, 'l'},
-  {"megabytes", no_argument, NULL, 'm'}, /* obsolescent */
+  {"megabytes", no_argument, NULL, MEGABYTES_OPTION}, /* obsolescent,  */
   {"portability", no_argument, NULL, 'P'},
   {"print-type", no_argument, NULL, 'T'},
   {"sync", no_argument, NULL, SYNC_OPTION},
@@ -142,75 +192,159 @@ static struct option const long_options[] =
   {NULL, 0, NULL, 0}
 };
 
-static void
-print_header (void)
+/* Replace problematic chars with '?'.
+   Since only control characters are currently considered,
+   this should work in all encodings.  */
+
+static char*
+hide_problematic_chars (char *cell)
 {
-  char buf[MAX (LONGEST_HUMAN_READABLE + 1, INT_BUFSIZE_BOUND (uintmax_t))];
-
-  if (print_type)
-    /* TRANSLATORS:
-       For best results (df header/column alignment), ensure that
-       your translation has the same length as the original.  */
-    fputs (_("Filesystem    Type"), stdout);
-  else
-    fputs (_("Filesystem        "), stdout);
-
-  if (inode_format)
-    /* TRANSLATORS:
-       For best results (df header/column alignment), ensure that
-       your translation has the same length as the original.
-       Also, each column name translation should end at the same
-       column as the corresponding original.  */
-    fputs (_("    Inodes   IUsed   IFree IUse%"), stdout);
-  else if (human_output_opts & human_autoscale)
+  char *p = cell;
+  while (*p)
     {
-      if (human_output_opts & human_base_1024)
-        fputs (_("    Size  Used Avail Use%"), stdout);
-      else
-        fputs (_("     Size   Used  Avail Use%"), stdout);
+      if (iscntrl (to_uchar (*p)))
+        *p = '?';
+      p++;
     }
-  else if (posix_format)
-    printf (_(" %s-blocks      Used Available Capacity"),
-            umaxtostr (output_block_size, buf));
-  else
+  return cell;
+}
+
+/* Dynamically allocate a row of pointers in TABLE, which
+   can then be accessed with standard 2D array notation.  */
+
+static void
+alloc_table_row (void)
+{
+  nrows++;
+  table = xnrealloc (table, nrows, sizeof (char *));
+  table[nrows-1] = xnmalloc (NFIELDS, sizeof (char *));
+}
+
+/* Output each cell in the table, accounting for the
+   alignment and max width of each column.  */
+
+static void
+print_table (void)
+{
+  size_t field, row;
+
+  for (row = 0; row < nrows; row ++)
     {
-      int opts = (human_suppress_point_zero
-                  | human_autoscale | human_SI
-                  | (human_output_opts
-                     & (human_group_digits | human_base_1024 | human_B)));
-
-      /* Prefer the base that makes the human-readable value more exact,
-         if there is a difference.  */
-
-      uintmax_t q1000 = output_block_size;
-      uintmax_t q1024 = output_block_size;
-      bool divisible_by_1000;
-      bool divisible_by_1024;
-
-      do
+      for (field = 0; field < NFIELDS; field++)
         {
-          divisible_by_1000 = q1000 % 1000 == 0;  q1000 /= 1000;
-          divisible_by_1024 = q1024 % 1024 == 0;  q1024 /= 1024;
+          size_t width = widths[field];
+          char *cell = table[row][field];
+          if (!cell) /* Missing type column, or mount point etc. */
+            continue;
+
+          /* Note the DEV_FIELD used to be displayed on it's own line
+             if (!posix_format && mbswidth (cell) > 20), but that
+             functionality is probably more problematic than helpful.  */
+          if (field != 0)
+            putchar (' ');
+          if (field == MNT_FIELD) /* The last one.  */
+            fputs (cell, stdout);
+          else
+            {
+              cell = ambsalign (cell, &width, alignments[field], 0);
+              /* When ambsalign fails, output unaligned data.  */
+              fputs (cell ? cell : table[row][field], stdout);
+              free (cell);
+            }
+          IF_LINT (free (table[row][field]));
         }
-      while (divisible_by_1000 & divisible_by_1024);
-
-      if (divisible_by_1000 < divisible_by_1024)
-        opts |= human_base_1024;
-      if (divisible_by_1024 < divisible_by_1000)
-        opts &= ~human_base_1024;
-      if (! (opts & human_base_1024))
-        opts |= human_B;
-
-      printf (_(" %4s-blocks      Used Available Use%%"),
-              human_readable (output_block_size, buf, opts, 1, 1));
+      putchar ('\n');
+      IF_LINT (free (table[row]));
     }
 
-  fputs (_(" Mounted on\n"), stdout);
+  IF_LINT (free (table));
+}
+
+/* Obtain the appropriate header entries.  */
+
+static void
+get_header (void)
+{
+  size_t field;
+
+  alloc_table_row ();
+
+  for (field = 0; field < NFIELDS; field++)
+    {
+      if (field == TYPE_FIELD && !print_type)
+        {
+          table[nrows-1][field] = NULL;
+          continue;
+        }
+
+      char *cell = NULL;
+      char const *header = _(headers[field][header_mode]);
+      if (!header)
+        header = _(headers[field][DEFAULT_MODE]);
+
+      if (header_mode == DEFAULT_MODE && field == TOTAL_FIELD)
+        {
+          char buf[LONGEST_HUMAN_READABLE + 1];
+
+          int opts = (human_suppress_point_zero
+                      | human_autoscale | human_SI
+                      | (human_output_opts
+                         & (human_group_digits | human_base_1024 | human_B)));
+
+          /* Prefer the base that makes the human-readable value more exact,
+             if there is a difference.  */
+
+          uintmax_t q1000 = output_block_size;
+          uintmax_t q1024 = output_block_size;
+          bool divisible_by_1000;
+          bool divisible_by_1024;
+
+          do
+            {
+              divisible_by_1000 = q1000 % 1000 == 0;  q1000 /= 1000;
+              divisible_by_1024 = q1024 % 1024 == 0;  q1024 /= 1024;
+            }
+          while (divisible_by_1000 & divisible_by_1024);
+
+          if (divisible_by_1000 < divisible_by_1024)
+            opts |= human_base_1024;
+          if (divisible_by_1024 < divisible_by_1000)
+            opts &= ~human_base_1024;
+          if (! (opts & human_base_1024))
+            opts |= human_B;
+
+          char *num = human_readable (output_block_size, buf, opts, 1, 1);
+
+          /* TRANSLATORS: this is the "1K-blocks" header in "df" output.  */
+          if (asprintf (&cell, _("%s-%s"), num, header) == -1)
+            cell = NULL;
+        }
+      else if (header_mode == POSIX_MODE && field == TOTAL_FIELD)
+        {
+          char buf[INT_BUFSIZE_BOUND (uintmax_t)];
+          char *num = umaxtostr (output_block_size, buf);
+
+          /* TRANSLATORS: this is the "1024-blocks" header in "df -P".  */
+          if (asprintf (&cell, _("%s-%s"), num, header) == -1)
+            cell = NULL;
+        }
+      else
+        cell = strdup (header);
+
+      if (!cell)
+        xalloc_die ();
+
+      hide_problematic_chars (cell);
+
+      table[nrows-1][field] = cell;
+
+      widths[field] = MAX (widths[field], mbswidth (cell, 0));
+    }
 }
 
 /* Is FSTYPE a type of file system that should be listed?  */
 
-static bool
+static bool _GL_ATTRIBUTE_PURE
 selected_fstype (const char *fstype)
 {
   const struct fs_type_list *fsp;
@@ -225,7 +359,7 @@ selected_fstype (const char *fstype)
 
 /* Is FSTYPE a type of file system that should be omitted?  */
 
-static bool
+static bool _GL_ATTRIBUTE_PURE
 excluded_fstype (const char *fstype)
 {
   const struct fs_type_list *fsp;
@@ -306,7 +440,18 @@ add_uint_with_neg_flag (uintmax_t *dest, bool *dest_neg,
     *dest = -*dest;
 }
 
-/* Display a space listing for the disk device with absolute file name DISK.
+/* Return true if S ends in a string that may be a 36-byte UUID,
+   i.e., of the form HHHHHHHH-HHHH-HHHH-HHHH-HHHHHHHHHHHH, where
+   each H is an upper or lower case hexadecimal digit.  */
+static bool _GL_ATTRIBUTE_PURE
+has_uuid_suffix (char const *s)
+{
+  size_t len = strlen (s);
+  return (36 < len
+          && strspn (s + len - 36, "-0123456789abcdefABCDEF") == 36);
+}
+
+/* Obtain a space listing for the disk device with absolute file name DISK.
    If MOUNT_POINT is non-NULL, it is the name of the root of the
    file system on DISK.
    If STAT_FILE is non-null, it is the name of a file within the file
@@ -317,19 +462,19 @@ add_uint_with_neg_flag (uintmax_t *dest, bool *dest_neg,
    If FSTYPE is non-NULL, it is the type of the file system on DISK.
    If MOUNT_POINT is non-NULL, then DISK may be NULL -- certain systems may
    not be able to produce statistics in this case.
-   ME_DUMMY and ME_REMOTE are the mount entry flags.  */
+   ME_DUMMY and ME_REMOTE are the mount entry flags.
+   Caller must set PROCESS_ALL to true when iterating over all entries, as
+   when df is invoked with no non-option argument.  See below for details.  */
 
 static void
-show_dev (char const *disk, char const *mount_point,
-          char const *stat_file, char const *fstype,
-          bool me_dummy, bool me_remote,
-          const struct fs_usage *force_fsu)
+get_dev (char const *disk, char const *mount_point,
+         char const *stat_file, char const *fstype,
+         bool me_dummy, bool me_remote,
+         const struct fs_usage *force_fsu,
+         bool process_all)
 {
   struct fs_usage fsu;
-  char buf[3][LONGEST_HUMAN_READABLE + 2];
-  int width;
-  int col1_adjustment = 0;
-  int use_width;
+  char buf[LONGEST_HUMAN_READABLE + 2];
   uintmax_t input_units;
   uintmax_t output_units;
   uintmax_t total;
@@ -339,6 +484,8 @@ show_dev (char const *disk, char const *mount_point,
   uintmax_t used;
   bool negate_used;
   double pct = -1;
+  char* cell;
+  size_t field;
 
   if (me_remote && show_local_fs)
     return;
@@ -370,40 +517,38 @@ show_dev (char const *disk, char const *mount_point,
 
   if (! file_systems_processed)
     {
-      file_systems_processed = true;
-      print_header ();
+      if (! force_fsu)
+        file_systems_processed = true;
+      get_header ();
     }
 
+  alloc_table_row ();
+
   if (! disk)
     disk = "-";			/* unknown */
+
+  char *dev_name = xstrdup (disk);
+  char *resolved_dev;
+
+  /* On some systems, dev_name is a long-named symlink like
+     /dev/disk/by-uuid/828fc648-9f30-43d8-a0b1-f7196a2edb66 pointing to a
+     much shorter and more useful name like /dev/sda1.  It may also look
+     like /dev/mapper/luks-828fc648-9f30-43d8-a0b1-f7196a2edb66 and point to
+     /dev/dm-0.  When process_all is true and dev_name is a symlink whose
+     name ends with a UUID use the resolved name instead.  */
+  if (process_all
+      && has_uuid_suffix (dev_name)
+      && (resolved_dev = canonicalize_filename_mode (dev_name, CAN_EXISTING)))
+    {
+      free (dev_name);
+      dev_name = resolved_dev;
+    }
+
   if (! fstype)
     fstype = "-";		/* unknown */
 
-  /* df.c reserved 5 positions for fstype,
-     but that does not suffice for type iso9660 */
-  if (print_type)
-    {
-      size_t disk_name_len = strlen (disk);
-      size_t fstype_len = strlen (fstype);
-      if (disk_name_len + fstype_len < 18)
-        printf ("%s%*s  ", disk, 18 - (int) disk_name_len, fstype);
-      else if (!posix_format)
-        printf ("%s\n%18s  ", disk, fstype);
-      else
-        printf ("%s %s", disk, fstype);
-    }
-  else
-    {
-      if (strlen (disk) > 20 && !posix_format)
-        printf ("%s\n%20s", disk, "");
-      else
-        printf ("%-20s", disk);
-    }
-
   if (inode_format)
     {
-      width = 7;
-      use_width = 5;
       input_units = output_units = 1;
       total = fsu.fsu_files;
       available = fsu.fsu_ffree;
@@ -417,22 +562,6 @@ show_dev (char const *disk, char const *mount_point,
     }
   else
     {
-      if (human_output_opts & human_autoscale)
-        width = 5 + ! (human_output_opts & human_base_1024);
-      else
-        {
-          width = 9;
-          if (posix_format)
-            {
-              uintmax_t b;
-              col1_adjustment = -3;
-              for (b = output_block_size; 9 < b; b /= 10)
-                col1_adjustment++;
-            }
-        }
-      use_width = ((posix_format
-                    && ! (human_output_opts & human_autoscale))
-                   ? 8 : 4);
       input_units = fsu.fsu_blocksize;
       output_units = output_block_size;
       total = fsu.fsu_blocks;
@@ -459,161 +588,113 @@ show_dev (char const *disk, char const *mount_point,
       negate_used = (total < available_to_root);
     }
 
-  printf (" %*s %*s %*s ",
-          width + col1_adjustment,
-          df_readable (false, total,
-                       buf[0], input_units, output_units),
-          width, df_readable (negate_used, used,
-                              buf[1], input_units, output_units),
-          width, df_readable (negate_available, available,
-                              buf[2], input_units, output_units));
-
-  if (! known_value (used) || ! known_value (available))
-    ;
-  else if (!negate_used
-           && used <= TYPE_MAXIMUM (uintmax_t) / 100
-           && used + available != 0
-           && (used + available < used) == negate_available)
+  for (field = 0; field < NFIELDS; field++)
     {
-      uintmax_t u100 = used * 100;
-      uintmax_t nonroot_total = used + available;
-      pct = u100 / nonroot_total + (u100 % nonroot_total != 0);
-    }
-  else
-    {
-      /* The calculation cannot be done easily with integer
-         arithmetic.  Fall back on floating point.  This can suffer
-         from minor rounding errors, but doing it exactly requires
-         multiple precision arithmetic, and it's not worth the
-         aggravation.  */
-      double u = negate_used ? - (double) - used : used;
-      double a = negate_available ? - (double) - available : available;
-      double nonroot_total = u + a;
-      if (nonroot_total)
+      switch (field)
         {
-          long int lipct = pct = u * 100 / nonroot_total;
-          double ipct = lipct;
+        case DEV_FIELD:
+          cell = dev_name;
+          break;
 
-          /* Like `pct = ceil (dpct);', but avoid ceil so that
-             the math library needn't be linked.  */
-          if (ipct - 1 < pct && pct <= ipct + 1)
-            pct = ipct + (ipct < pct);
-        }
-    }
+        case TYPE_FIELD:
+          cell = print_type ? xstrdup (fstype) : NULL;
+          break;
 
-  if (0 <= pct)
-    printf ("%*.0f%%", use_width - 1, pct);
-  else
-    printf ("%*s", use_width, "- ");
+        case TOTAL_FIELD:
+          cell = xstrdup (df_readable (false, total, buf,
+                                       input_units, output_units));
+          break;
+        case USED_FIELD:
+          cell = xstrdup (df_readable (negate_used, used, buf,
+                                       input_units, output_units));
+          break;
+        case FREE_FIELD:
+          cell = xstrdup (df_readable (negate_available, available, buf,
+                                       input_units, output_units));
+          break;
 
-  if (mount_point)
-    {
+        case PCENT_FIELD:
+          if (! known_value (used) || ! known_value (available))
+            ;
+          else if (!negate_used
+                   && used <= TYPE_MAXIMUM (uintmax_t) / 100
+                   && used + available != 0
+                   && (used + available < used) == negate_available)
+            {
+              uintmax_t u100 = used * 100;
+              uintmax_t nonroot_total = used + available;
+              pct = u100 / nonroot_total + (u100 % nonroot_total != 0);
+            }
+          else
+            {
+              /* The calculation cannot be done easily with integer
+                 arithmetic.  Fall back on floating point.  This can suffer
+                 from minor rounding errors, but doing it exactly requires
+                 multiple precision arithmetic, and it's not worth the
+                 aggravation.  */
+              double u = negate_used ? - (double) - used : used;
+              double a = negate_available ? - (double) - available : available;
+              double nonroot_total = u + a;
+              if (nonroot_total)
+                {
+                  long int lipct = pct = u * 100 / nonroot_total;
+                  double ipct = lipct;
+
+                  /* Like 'pct = ceil (dpct);', but avoid ceil so that
+                     the math library needn't be linked.  */
+                  if (ipct - 1 < pct && pct <= ipct + 1)
+                    pct = ipct + (ipct < pct);
+                }
+            }
+
+          if (0 <= pct)
+            {
+              if (asprintf (&cell, "%.0f%%", pct) == -1)
+                cell = NULL;
+            }
+          else
+            cell = strdup ("-");
+
+          if (!cell)
+            xalloc_die ();
+
+          break;
+
+        case MNT_FIELD:
+          if (mount_point)
+            {
 #ifdef HIDE_AUTOMOUNT_PREFIX
-      /* Don't print the first directory name in MOUNT_POINT if it's an
-         artifact of an automounter.  This is a bit too aggressive to be
-         the default.  */
-      if (strncmp ("/auto/", mount_point, 6) == 0)
-        mount_point += 5;
-      else if (strncmp ("/tmp_mnt/", mount_point, 9) == 0)
-        mount_point += 8;
+              /* Don't print the first directory name in MOUNT_POINT if it's an
+                 artifact of an automounter.  This is a bit too aggressive to be
+                 the default.  */
+              if (STRNCMP_LIT (mount_point, "/auto/") == 0)
+                mount_point += 5;
+              else if (STRNCMP_LIT (mount_point, "/tmp_mnt/") == 0)
+                mount_point += 8;
 #endif
-      printf (" %s", mount_point);
-    }
-  putchar ('\n');
-}
+              cell = xstrdup (mount_point);
+            }
+          else
+            cell = NULL;
+          break;
 
-/* Return the root mountpoint of the file system on which FILE exists, in
-   malloced storage.  FILE_STAT should be the result of stating FILE.
-   Give a diagnostic and return NULL if unable to determine the mount point.
-   Exit if unable to restore current working directory.  */
-static char *
-find_mount_point (const char *file, const struct stat *file_stat)
-{
-  struct saved_cwd cwd;
-  struct stat last_stat;
-  char *mp = NULL;		/* The malloced mount point.  */
-
-  if (save_cwd (&cwd) != 0)
-    {
-      error (0, errno, _("cannot get current directory"));
-      return NULL;
-    }
-
-  if (S_ISDIR (file_stat->st_mode))
-    /* FILE is a directory, so just chdir there directly.  */
-    {
-      last_stat = *file_stat;
-      if (chdir (file) < 0)
-        {
-          error (0, errno, _("cannot change to directory %s"), quote (file));
-          return NULL;
-        }
-    }
-  else
-    /* FILE is some other kind of file; use its directory.  */
-    {
-      char *xdir = dir_name (file);
-      char *dir;
-      ASSIGN_STRDUPA (dir, xdir);
-      free (xdir);
-
-      if (chdir (dir) < 0)
-        {
-          error (0, errno, _("cannot change to directory %s"), quote (dir));
-          return NULL;
+        default:
+          assert (!"unhandled field");
         }
 
-      if (stat (".", &last_stat) < 0)
+      if (cell)
         {
-          error (0, errno, _("cannot stat current directory (now %s)"),
-                 quote (dir));
-          goto done;
+          hide_problematic_chars (cell);
+          widths[field] = MAX (widths[field], mbswidth (cell, 0));
         }
+      table[nrows-1][field] = cell;
     }
-
-  /* Now walk up FILE's parents until we find another file system or /,
-     chdiring as we go.  LAST_STAT holds stat information for the last place
-     we visited.  */
-  for (;;)
-    {
-      struct stat st;
-      if (stat ("..", &st) < 0)
-        {
-          error (0, errno, _("cannot stat %s"), quote (".."));
-          goto done;
-        }
-      if (st.st_dev != last_stat.st_dev || st.st_ino == last_stat.st_ino)
-        /* cwd is the mount point.  */
-        break;
-      if (chdir ("..") < 0)
-        {
-          error (0, errno, _("cannot change to directory %s"), quote (".."));
-          goto done;
-        }
-      last_stat = st;
-    }
-
-  /* Finally reached a mount point, see what it's called.  */
-  mp = xgetcwd ();
-
-done:
-  /* Restore the original cwd.  */
-  {
-    int save_errno = errno;
-    if (restore_cwd (&cwd) != 0)
-      error (EXIT_FAILURE, errno,
-             _("failed to return to initial working directory"));
-    free_cwd (&cwd);
-    errno = save_errno;
-  }
-
-  return mp;
 }
 
 /* If DISK corresponds to a mount point, show its usage
    and return true.  Otherwise, return false.  */
 static bool
-show_disk (char const *disk)
+get_disk (char const *disk)
 {
   struct mount_entry const *me;
   struct mount_entry const *best_match = NULL;
@@ -624,9 +705,9 @@ show_disk (char const *disk)
 
   if (best_match)
     {
-      show_dev (best_match->me_devname, best_match->me_mountdir, NULL,
-                best_match->me_type, best_match->me_dummy,
-                best_match->me_remote, NULL);
+      get_dev (best_match->me_devname, best_match->me_mountdir, NULL,
+               best_match->me_type, best_match->me_dummy,
+               best_match->me_remote, NULL, false);
       return true;
     }
 
@@ -635,62 +716,43 @@ show_disk (char const *disk)
 
 /* Figure out which device file or directory POINT is mounted on
    and show its disk usage.
-   STATP must be the result of `stat (POINT, STATP)'.  */
+   STATP must be the result of 'stat (POINT, STATP)'.  */
 static void
-show_point (const char *point, const struct stat *statp)
+get_point (const char *point, const struct stat *statp)
 {
   struct stat disk_stats;
   struct mount_entry *me;
   struct mount_entry const *best_match = NULL;
 
-  /* If POINT is an absolute file name, see if we can find the
-     mount point without performing any extra stat calls at all.  */
-  if (*point == '/')
-    {
-      /* Find the best match: prefer non-dummies, and then prefer the
-         last match if there are ties.  */
-
-      for (me = mount_list; me; me = me->me_next)
-        if (STREQ (me->me_mountdir, point) && !STREQ (me->me_type, "lofs")
-            && (!best_match || best_match->me_dummy || !me->me_dummy))
-          best_match = me;
-    }
-
   /* Calculate the real absolute file name for POINT, and use that to find
      the mount point.  This avoids statting unavailable mount points,
      which can hang df.  */
-  if (! best_match)
+  char *resolved = canonicalize_file_name (point);
+  if (resolved && resolved[0] == '/')
     {
-      char *resolved = canonicalize_file_name (point);
+      size_t resolved_len = strlen (resolved);
+      size_t best_match_len = 0;
 
-      if (resolved && resolved[0] == '/')
+      for (me = mount_list; me; me = me->me_next)
+      if (!STREQ (me->me_type, "lofs")
+          && (!best_match || best_match->me_dummy || !me->me_dummy))
         {
-          size_t resolved_len = strlen (resolved);
-          size_t best_match_len = 0;
-
-          for (me = mount_list; me; me = me->me_next)
-            if (!STREQ (me->me_type, "lofs")
-                && (!best_match || best_match->me_dummy || !me->me_dummy))
-              {
-                size_t len = strlen (me->me_mountdir);
-                if (best_match_len <= len && len <= resolved_len
-                    && (len == 1 /* root file system */
-                        || ((len == resolved_len || resolved[len] == '/')
-                            && strncmp (me->me_mountdir, resolved, len) == 0)))
-                  {
-                    best_match = me;
-                    best_match_len = len;
-                  }
-              }
+          size_t len = strlen (me->me_mountdir);
+          if (best_match_len <= len && len <= resolved_len
+              && (len == 1 /* root file system */
+                  || ((len == resolved_len || resolved[len] == '/')
+                      && STREQ_LEN (me->me_mountdir, resolved, len))))
+            {
+              best_match = me;
+              best_match_len = len;
+            }
         }
-
-      free (resolved);
-
-      if (best_match
-          && (stat (best_match->me_mountdir, &disk_stats) != 0
-              || disk_stats.st_dev != statp->st_dev))
-        best_match = NULL;
     }
+  free (resolved);
+  if (best_match
+      && (stat (best_match->me_mountdir, &disk_stats) != 0
+          || disk_stats.st_dev != statp->st_dev))
+    best_match = NULL;
 
   if (! best_match)
     for (me = mount_list; me; me = me->me_next)
@@ -729,9 +791,9 @@ show_point (const char *point, const struct stat *statp)
       }
 
   if (best_match)
-    show_dev (best_match->me_devname, best_match->me_mountdir, point,
-              best_match->me_type, best_match->me_dummy, best_match->me_remote,
-              NULL);
+    get_dev (best_match->me_devname, best_match->me_mountdir, point,
+             best_match->me_type, best_match->me_dummy, best_match->me_remote,
+             NULL, false);
   else
     {
       /* We couldn't find the mount entry corresponding to POINT.  Go ahead and
@@ -742,36 +804,36 @@ show_point (const char *point, const struct stat *statp)
       char *mp = find_mount_point (point, statp);
       if (mp)
         {
-          show_dev (NULL, mp, NULL, NULL, false, false, NULL);
+          get_dev (NULL, mp, NULL, NULL, false, false, NULL, false);
           free (mp);
         }
     }
 }
 
 /* Determine what kind of node NAME is and show the disk usage
-   for it.  STATP is the results of `stat' on NAME.  */
+   for it.  STATP is the results of 'stat' on NAME.  */
 
 static void
-show_entry (char const *name, struct stat const *statp)
+get_entry (char const *name, struct stat const *statp)
 {
   if ((S_ISBLK (statp->st_mode) || S_ISCHR (statp->st_mode))
-      && show_disk (name))
+      && get_disk (name))
     return;
 
-  show_point (name, statp);
+  get_point (name, statp);
 }
 
 /* Show all mounted file systems, except perhaps those that are of
    an unselected type or are empty. */
 
 static void
-show_all_entries (void)
+get_all_entries (void)
 {
   struct mount_entry *me;
 
   for (me = mount_list; me; me = me->me_next)
-    show_dev (me->me_devname, me->me_mountdir, NULL, me->me_type,
-              me->me_dummy, me->me_remote, NULL);
+    get_dev (me->me_devname, me->me_mountdir, NULL, me->me_type,
+             me->me_dummy, me->me_remote, NULL, true);
 }
 
 /* Add FSTYPE to the list of file system types to display. */
@@ -804,8 +866,7 @@ void
 usage (int status)
 {
   if (status != EXIT_SUCCESS)
-    fprintf (stderr, _("Try `%s --help' for more information.\n"),
-             program_name);
+    emit_try_help ();
   else
     {
       printf (_("Usage: %s [OPTION]... [FILE]...\n"), program_name);
@@ -819,16 +880,20 @@ Mandatory arguments to long options are mandatory for short options too.\n\
 "), stdout);
       fputs (_("\
   -a, --all             include dummy file systems\n\
-  -B, --block-size=SIZE  use SIZE-byte blocks\n\
+  -B, --block-size=SIZE  scale sizes by SIZE before printing them.  E.g.,\n\
+                           '-BM' prints sizes in units of 1,048,576 bytes.\n\
+                           See SIZE format below.\n\
       --total           produce a grand total\n\
-  -h, --human-readable  print sizes in human readable format (e.g., 1K 234M 2G)\n\
+  -h, --human-readable  print sizes in human readable format (e.g., 1K 234M 2G)\
+\n\
   -H, --si              likewise, but use powers of 1000 not 1024\n\
 "), stdout);
       fputs (_("\
   -i, --inodes          list inode information instead of block usage\n\
   -k                    like --block-size=1K\n\
   -l, --local           limit listing to local file systems\n\
-      --no-sync         do not invoke sync before getting usage info (default)\n\
+      --no-sync         do not invoke sync before getting usage info (default)\
+\n\
 "), stdout);
       fputs (_("\
   -P, --portability     use the POSIX output format\n\
@@ -850,7 +915,7 @@ Mandatory arguments to long options are mandatory for short options too.\n\
 int
 main (int argc, char **argv)
 {
-  struct stat *stats IF_LINT (= 0);
+  struct stat *stats IF_LINT ( = 0);
 
   initialize_main (&argc, &argv);
   set_program_name (argv[0]);
@@ -873,7 +938,7 @@ main (int argc, char **argv)
   print_grand_total = false;
   grand_fsu.fsu_blocksize = 1;
 
-  for (;;)
+  while (true)
     {
       int oi = -1;
       int c = getopt_long (argc, argv, "aB:iF:hHklmPTt:vx:", long_options,
@@ -912,7 +977,14 @@ main (int argc, char **argv)
         case 'l':
           show_local_fs = true;
           break;
-        case 'm': /* obsolescent */
+        case MEGABYTES_OPTION:
+          /* Distinguish between the long and the short option.
+             As we want to remove the long option soon,
+             give a warning when the long form is used.  */
+          error (0, 0, "%s%s", _("warning: "),
+            _("long option '--megabytes' is deprecated"
+              " and will soon be removed"));
+        case 'm': /* obsolescent, exists for BSD compatibility */
           human_output_opts = 0;
           output_block_size = 1024 * 1024;
           break;
@@ -966,6 +1038,13 @@ main (int argc, char **argv)
                        &human_output_opts, &output_block_size);
     }
 
+  if (inode_format)
+    header_mode = INODES_MODE;
+  else if (human_output_opts & human_autoscale)
+    header_mode = HUMAN_MODE;
+  else if (posix_format)
+    header_mode = POSIX_MODE;
+
   /* Fail if the same file system type was both selected and excluded.  */
   {
     bool match = false;
@@ -1023,10 +1102,19 @@ main (int argc, char **argv)
   if (mount_list == NULL)
     {
       /* Couldn't read the table of mounted file systems.
-         Fail if df was invoked with no file name arguments;
-         Otherwise, merely give a warning and proceed.  */
-      int status =          (optind < argc ? 0 : EXIT_FAILURE);
-      const char *warning = (optind < argc ? _("Warning: ") : "");
+         Fail if df was invoked with no file name arguments,
+         or when either of -a, -l, -t or -x is used with file name
+         arguments. Otherwise, merely give a warning and proceed.  */
+      int status = 0;
+      if ( ! (optind < argc)
+           || (show_all_fs
+               || show_local_fs
+               || fs_select_list != NULL
+               || fs_exclude_list != NULL))
+        {
+          status = EXIT_FAILURE;
+        }
+      const char *warning = (status == 0 ? _("Warning: ") : "");
       error (status, errno, "%s%s", warning,
              _("cannot read table of mounted file systems"));
     }
@@ -1043,19 +1131,23 @@ main (int argc, char **argv)
 
       for (i = optind; i < argc; ++i)
         if (argv[i])
-          show_entry (argv[i], &stats[i - optind]);
+          get_entry (argv[i], &stats[i - optind]);
     }
   else
-    show_all_entries ();
+    get_all_entries ();
 
-  if (print_grand_total)
+  if (print_grand_total && file_systems_processed)
     {
       if (inode_format)
         grand_fsu.fsu_blocks = 1;
-      show_dev ("total", NULL, NULL, NULL, false, false, &grand_fsu);
+      get_dev ("total", NULL, NULL, NULL, false, false, &grand_fsu, false);
     }
 
-  if (! file_systems_processed)
+  print_table ();
+
+  /* Print the "no FS processed" diagnostic only if there was no preceding
+     diagnostic, e.g., if all have been excluded.  */
+  if (exit_status == EXIT_SUCCESS && ! file_systems_processed)
     error (EXIT_FAILURE, 0, _("no file systems processed"));
 
   exit (exit_status);
